/**
 * PETALS - PETALS Services Platform.
 * Copyright (c) 2007 EBM Websourcing, http://www.ebmwebsourcing.com/
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * -------------------------------------------------------------------------
 * JBIDescriptorBuilder - Initial Developper : ofabre - EBM Websourcing
 * -------------------------------------------------------------------------
 */
package org.objectweb.petals.jbi.descriptor;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.net.URL;
import java.util.ArrayList;
import java.util.List;

import javax.xml.XMLConstants;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;
import javax.xml.validation.Schema;
import javax.xml.validation.SchemaFactory;
import javax.xml.validation.Validator;

import org.apache.commons.io.IOUtils;
import org.objectweb.petals.jbi.descriptor.util.XMLUtil;
import org.w3c.dom.Document;
import org.w3c.dom.DocumentFragment;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;

/**
 * It's the Petals JBIDescriptor builder
 *
 * @author ofabre - EBM Websourcing
 *
 */
public class JBIDescriptorBuilder {

	/**
	 * Singleton instance
	 */
	private static JBIDescriptorBuilder ref;

	public static final String JBI_XSD = "jbi.xsd";

	/**
	 * Standard location to the JBI Descriptor resource.
	 */
	public static final String JBI_DESCRIPTOR_RESOURCE = "META-INF"
			+ File.separator + "jbi.xml";

	/**
	 * The JBI Descriptor XML Schema Resource Location ("schema/jbi.xsd").
	 */
	public static final String JBI_SCHEMA_RESOURCE = "schema" + File.separator
			+ JBI_XSD;

	private static final String BOOTSTRAP_CLASS_NAME_ELEM_NAME = "bootstrap-class-name";

	private static final String BOOTSTRAP_CLASS_PATH_ELEM_NAME = "bootstrap-class-path";

	private static final String COMPONENT_CLASS_NAME_ELEM_NAME = "component-class-name";

	private static final String COMPONENT_CLASS_PATH_ELEM_NAME = "component-class-path";

	private static final String ENDPOINT_ATTR_NAME = "endpoint-name";

	private static final String IDENTIFICATION_ELEM_NAME = "identification";

	private static final String INTERFACE_ATTR_NAME = "interface-name";

	/**
	 * Schema file
	 */
	private static Schema schema = null;

	private static final String SERVICE_ATTR_NAME = "service-name";

	private static final String SHARED_LIBRARY_ELEM_NAME = "shared-library";

	private static final String VERSION_ATTR_NAME = "version";

	private JBIDescriptorBuilder() {
		// nothing to do
	}

	/**
	 *
	 * @return a single instance of the {@link JBIDescriptorBuilder}
	 * @throws JBIDescriptorException
	 *             if it can't correctly load jbi.xsd schema file
	 */
	public static JBIDescriptorBuilder getInstance() throws JBIDescriptorException {
		if (ref == null) {
			ref = new JBIDescriptorBuilder();

				ref.loadJBISchema();

		}
		return ref;
	}

	/**
	 * Extension nodes are nodes which names are different from those specified
	 * by JBI. This method load these extensions.
	 *
	 * @param node
	 *            an extension xml node
	 * @throws JBIDescriptorException
	 */
	private void loadExtensions(final Extensions extensions, final Node node)
			throws JBIDescriptorException {

		// Add the extension node to the extensions documentfragment containing
		// all extensions
		DocumentFragment fragment = extensions.getDocumentFragment();
		if (fragment == null) {
			fragment = node.getOwnerDocument().createDocumentFragment();
			extensions.setDocumentFragment(fragment);
		}
		fragment.appendChild(node);
	}

	/**
	 * Returns a String representation of a JBI descriptor file
	 *
	 * @param jbiXmlUri
	 *            the JBI Descriptor file <code>URI</code>. Must not be null
	 *            and must point on a real jbi.xml file
	 * @return a <code>String</code> representation of the JBI Descriptor
	 * @throws JBIDescriptorException
	 */
	public String getJbiDescriptorAsString(URI jbiXmlUri)
			throws JBIDescriptorException {
		if (jbiXmlUri == null) {
			throw new IllegalArgumentException("jbiXmlUri must not be null");
		}
		File descriptorFile = new File(jbiXmlUri);
		InputStream is = null;
		if (descriptorFile.isFile()) {
			try {
				is = new FileInputStream(descriptorFile);
			} catch (FileNotFoundException e) {
				throw new JBIDescriptorException(
						"Can't find jbi descriptor file: "
								+ jbiXmlUri.toString(), e);
			}
		} else {
			throw new IllegalArgumentException(
					"Given URI must point on a real jbi.xml file :" + jbiXmlUri);
		}
		return getJbiDescriptorAsString(is);
	}

	/**
	 * Returns a String representation of a JBI descriptor file
	 *
	 * @param jbiXmlUri
	 *            the JBI Descriptor file <code>URI</code>. Must not be null.
	 * @return a <code>String</code> representation of the JBI Descriptor
	 * @throws JBIDescriptorException
	 */
	public String getJbiDescriptorAsString(InputStream jbiXmlStream)
			throws JBIDescriptorException {
		if (jbiXmlStream == null) {
			throw new IllegalArgumentException("jbiXmlStream must not be null");
		}
		String ret = null;
		ByteArrayOutputStream os = null;
		try {
			os = new ByteArrayOutputStream();
			IOUtils.copy(jbiXmlStream, os);
			ret = os.toString();
		} catch (IOException e) {
			throw new JBIDescriptorException("Can't read descriptor as String",
					e);
		}
		return ret;
	}

	/**
	 * Validate the jbi.xml file denoted by the given {@link URI} and create a
	 * {@link JBIDescriptor} from it
	 *
	 * @param jbiXmlUri
	 * @return {@link JBIDescriptor} representation of the jbi.xml file
	 * @throws JBIDescriptorException
	 */
	public JBIDescriptor validateAndBuildDescriptor(URI jbiXmlUri)
			throws JBIDescriptorException {
		validateJbiXml(jbiXmlUri);
		Document document = parseJbiXml(jbiXmlUri);
		return createJBIDescriptor(document);
	}

	/**
	 * Validate the jbi.xml file contained in the {@link InputStream} and create
	 * a {@link JBIDescriptor} from it
	 *
	 * @param jbiXmlStream
	 * @return {@link JBIDescriptor} representation of the jbi.xml file
	 * @throws JBIDescriptorException
	 */
	public JBIDescriptor validateAndBuildDescriptor(InputStream jbiXmlStream)
			throws JBIDescriptorException {
		validateJbiXml(jbiXmlStream);
		Document document = parseJbiXml(jbiXmlStream);
		return createJBIDescriptor(document);
	}

	/**
	 * Create a {@link JBIDescriptor} from jbi.xml file without validation
	 *
	 * @param jbiXmlUri
	 * @return {@link JBIDescriptor} representation of the jbi.xml file
	 * @throws JBIDescriptorException
	 */
	public JBIDescriptor buildDescriptor(URI jbiXmlUri)
			throws JBIDescriptorException {
		Document document = parseJbiXml(jbiXmlUri);
		return createJBIDescriptor(document);
	}

	/**
	 * Create a {@link JBIDescriptor} from jbi.xml file without validation
	 *
	 * @param jbiXmlStream
	 * @return {@link JBIDescriptor} representation of the jbi.xml file
	 * @throws JBIDescriptorException
	 */
	public JBIDescriptor buildDescriptor(InputStream jbiXmlStream)
			throws JBIDescriptorException {
		Document document = parseJbiXml(jbiXmlStream);
		return createJBIDescriptor(document);
	}

	/**
	 *
	 * @param sourceURI
	 *            {@link URI} of a jbi.xml descriptor file
	 * @throws JBIDescriptorException
	 *             if an error occurs during jbi.xml validation
	 */
	public void validateJbiXml(final URI sourceURI)
			throws JBIDescriptorException {
		validateJbiXml(new StreamSource(new File(sourceURI)));
	}

	/**
	 *
	 * @param sourceStream
	 *            {@link InputStream} containing jbi.xml descriptor file
	 * @throws JBIDescriptorException
	 *             if an error occurs during jbi.xml validation
	 */
	public void validateJbiXml(final InputStream sourceStream)
			throws JBIDescriptorException {
		validateJbiXml(new StreamSource(sourceStream));
	}

	/**
	 *
	 * @param source
	 *            a {@link Source} containing jbi.xml descriptor file stream
	 * @throws JBIDescriptorException
	 *             if an error occurs during jbi.xml validation
	 */
	public void validateJbiXml(final Source source)
			throws JBIDescriptorException {
		if (schema != null) {
			// create a Validator instance which is be used to validate a
			// document
			Validator validator = schema.newValidator();

			// validate the DOM tree
			try {
				validator.validate(source);
			} catch (SAXException e) {
				throw new JBIDescriptorException(
						"Error validating jbi.xml file", e);
			} catch (IOException e) {
				throw new JBIDescriptorException(
						"Error validating jbi.xml file", e);
			}
		}
	}

	/**
	 * Parse a jbi.xml file to create a {@link Document} representation
	 *
	 * @param sourceURI
	 *            an {@link URI} pointing on a jbi.xml file
	 * @return a {@link Document} representation of the jbi.xml file
	 * @throws JBIDescriptorException
	 *             if it can't create DocumentBuilder or if an error occurs
	 *             during file parsing
	 */
	public Document parseJbiXml(final URI sourceURI)
			throws JBIDescriptorException {
		try {
			return parseJbiXml(new FileInputStream(new File(sourceURI)));
		} catch (FileNotFoundException e) {
			throw new JBIDescriptorException("Can't find jbi.xml file", e);
		}
	}

	/**
	 * Parse a jbi.xml file to create a {@link Document} representation
	 *
	 * @param source
	 *            an {@link InputStream} containing the jbi.xml file stream
	 * @return a {@link Document} representation of the jbi.xml file
	 * @throws JBIDescriptorException
	 *             if it can't create DocumentBuilder or if an error occurs
	 *             during file parsing
	 */
	public Document parseJbiXml(final InputStream source)
			throws JBIDescriptorException {

		DocumentBuilder docBuilder = createDocumentBuilder();

		Document document = null;

		try {
			document = docBuilder.parse(source);
		} catch (SAXException saxe) {
			throw new JBIDescriptorException(
					"JBI descriptor is not well formed XML", saxe);
		} catch (IOException ioe) {
			throw new JBIDescriptorException("Can't read JBI descriptor", ioe);
		}
		return document;
	}

	private DocumentBuilder createDocumentBuilder()
			throws JBIDescriptorException {
		DocumentBuilderFactory docBuilderFactory = DocumentBuilderFactory
				.newInstance();
		docBuilderFactory.setNamespaceAware(true);

		DocumentBuilder docBuilder = null;

		try {
			docBuilder = docBuilderFactory.newDocumentBuilder();
		} catch (ParserConfigurationException pce) {
			throw new JBIDescriptorException("Bad XML parser configuration",
					pce);
		}
		return docBuilder;
	}

	/**
	 * Parse the XML document and load values to the Descriptor object
	 * representation.
	 *
	 * @param document
	 * @return a populated Descriptor instance with values loaded from the XML
	 *         file. May return <code>null</code> if an error occur while
	 *         navigating on XML tree.
	 * @throws JBIDescriptorException
	 */
	public JBIDescriptor createJBIDescriptor(final Document document)
			throws JBIDescriptorException {

		JBIDescriptor descriptor = new JBIDescriptor();

		// get root element
		Element jbi = document.getDocumentElement();

		// extract JBI version attribute from the root element
		String versionAttr;
		try {
			versionAttr = XMLUtil.getRequiredAttributeValue(jbi,
					VERSION_ATTR_NAME);
		} catch (NullPointerException e) {
			throw new JBIDescriptorException(e);
		}

		descriptor.setVersion(Double.parseDouble(versionAttr));

		List<Node> children = XMLUtil.getNodeChildren(jbi);
		for (Node child : children) {
			if ("component".equals(child.getNodeName())) {
				descriptor.setComponent(loadComponentXml(child));
			} else if (SHARED_LIBRARY_ELEM_NAME.equals(child.getNodeName())) {
				descriptor.setSharedLibrary(loadSharedLibraryXml(child));
			} else if ("service-assembly".equals(child.getNodeName())) {
				descriptor.setServiceAssembly(loadServiceAssemblyXml(child));
			} else if ("services".equals(child.getNodeName())) {
				descriptor.setServices(loadServicesXml(child));
			}
		}
		return descriptor;
	}

	/**
	 * @param element
	 * @throws JBIDescriptorException
	 * @throws
	 */
	private ComponentDescription loadComponentXml(final Node production)
			throws JBIDescriptorException {
		ComponentDescription component = new ComponentDescription();

		// component attributes
		try {
			component.setType(XMLUtil.getRequiredAttributeValue(production,
					"type"));
		} catch (NullPointerException e) {
			throw new JBIDescriptorException(e);
		}
		component.setComponentClassLoaderDelegation(XMLUtil.getAttributeValue(
				production, "component-class-loader-delegation"));
		component.setBootstrapClassLoaderDelegation(XMLUtil.getAttributeValue(
				production, "bootstrap-class-loader-delegation"));

		Extensions extensions = new Extensions();
		List<SharedLibraryList> slList = new ArrayList<SharedLibraryList>();
		// component sub nodes
		List<Node> children = XMLUtil.getNodeChildren(production);
		for (Node child : children) {
			if (IDENTIFICATION_ELEM_NAME.equals(child.getNodeName())) {
				component.setIdentification(loadIdentificationXml(child));
			} else if (COMPONENT_CLASS_NAME_ELEM_NAME.equals(child
					.getNodeName())) {
				component.setComponentClassName(XMLUtil.getTextContent(child));
			} else if (COMPONENT_CLASS_PATH_ELEM_NAME.equals(child
					.getNodeName())) {
				component.setComponentClassPath(XMLUtil.getTextContents(child
						.getChildNodes()));
			} else if (BOOTSTRAP_CLASS_NAME_ELEM_NAME.equals(child
					.getNodeName())) {
				component.setBootstrapClassName(XMLUtil.getTextContent(child));
			} else if (BOOTSTRAP_CLASS_PATH_ELEM_NAME.equals(child
					.getNodeName())) {
				component.setBootstrapClassPath(XMLUtil.getTextContents(child
						.getChildNodes()));
			} else if (SHARED_LIBRARY_ELEM_NAME.equals(child.getNodeName())) {
				loadSharedLibraryListXml(slList, child);
			} else {
				loadExtensions(extensions, child);
			}
		}
		// set extensions
		component.setExtensions(extensions);
		// set shared libs
		component.setSharedLibraryList(slList);

		return component;
	}

	/**
	 * @param connectionsTree
	 * @return
	 * @throws JBIDescriptorException
	 */
	private List<Connection> loadConnectionsXml(final Node connectionsTree)
			throws JBIDescriptorException {

		List<Connection> connections = new ArrayList<Connection>();

		if (connectionsTree != null) {
			// get "connections/connection" Node List
			List<Node> children = XMLUtil.getNodeChildren(connectionsTree);
			for (Node child : children) {
				if ("connection".equals(child.getNodeName())) {
					connections.add(loadConnectionXml(child));
				}
			}
		}
		return connections;
	}

	private Connection loadConnectionXml(final Node node)
			throws JBIDescriptorException {

		// connection object
		Connection newItem = new Connection();

		List<Node> children = XMLUtil.getNodeChildren(node);
		for (Node child : children) {
			if ("consumer".equals(child.getNodeName())) {
				/*
				 * TODO deal with cases where there is only the interface name
				 * and no endpoint name and service name
				 * newItem.setConsumerInterfaceName(extractXmlAttributeQName(consumer,
				 * "interface-name"));
				 */

				newItem.setConsumerServiceName(XMLUtil
						.extractXmlAttributeQName(child, SERVICE_ATTR_NAME));

				newItem.setConsumerEndpointName(XMLUtil.getAttributeValue(
						child, ENDPOINT_ATTR_NAME));
			} else if ("provider".equals(child.getNodeName())) {
				try {
					newItem.setProviderServiceName(XMLUtil
							.extractRequiredXmlAttributeQName(child,
									SERVICE_ATTR_NAME));
					newItem.setProviderEndpointName(XMLUtil
							.getRequiredAttributeValue(child,
									ENDPOINT_ATTR_NAME));
				} catch (NullPointerException e) {
					throw new JBIDescriptorException(e);
				}

			}
		}
		return newItem;
	}

	/**
	 * @param list
	 * @return
	 * @throws JBIDescriptorException
	 */
	private Consumes loadConsumesXml(final Node node)
			throws JBIDescriptorException {
		Consumes newItem = new Consumes();

		// consumes attributes
		try {
			newItem.setInterfaceName(XMLUtil.extractRequiredXmlAttributeQName(
					node, INTERFACE_ATTR_NAME));
		} catch (NullPointerException e) {
			throw new JBIDescriptorException(e);
		}

		newItem.setServiceName(XMLUtil.extractXmlAttributeQName(node,
				SERVICE_ATTR_NAME));

		newItem.setEndpointName(XMLUtil.getAttributeValue(node,
				ENDPOINT_ATTR_NAME));

		newItem.setLinkType(XMLUtil.getAttributeValue(node, "link-type"));

		// consumes extensions
		Extensions extensions = new Extensions();
		List<Node> children = XMLUtil.getNodeChildren(node);
		for (Node child : children) {
			loadExtensions(extensions, child);
		}
		newItem.setExtensions(extensions);

		return newItem;
	}

	private Identification loadIdentificationXml(final Node identificationNode)
			throws JBIDescriptorException {
		Identification identification = new Identification();

		Extensions extensions = new Extensions();
		// identification sub nodes
		List<Node> children = XMLUtil.getNodeChildren(identificationNode);
		for (Node child : children) {
			if ("name".equals(child.getNodeName())) {
				identification.setName(XMLUtil.getTextContent(child));
			} else if ("description".equals(child.getNodeName())) {
				identification.setDescription(XMLUtil.getTextContent(child));
			} else {
				loadExtensions(extensions, child);
			}
		}
		// set extensions
		identification.setExtensions(extensions);

		return identification;
	}

	/**
	 * @param list
	 * @return
	 * @throws JBIDescriptorException
	 */
	private Provides loadProvidesXml(final Node node)
			throws JBIDescriptorException {
		Provides newItem = new Provides();

		// Provides attributes
		try {
			newItem.setInterfaceName(XMLUtil.extractRequiredXmlAttributeQName(
					node, INTERFACE_ATTR_NAME));

			newItem.setServiceName(XMLUtil.extractRequiredXmlAttributeQName(
					node, SERVICE_ATTR_NAME));

			newItem.setEndpointName(XMLUtil.getRequiredAttributeValue(node,
					ENDPOINT_ATTR_NAME));
		} catch (NullPointerException e) {
			throw new JBIDescriptorException(e);
		}
		// provides extensions
		Extensions extensions = new Extensions();
		List<Node> children = XMLUtil.getNodeChildren(node);
		for (Node child : children) {
			loadExtensions(extensions, child);
		}
		newItem.setExtensions(extensions);

		return newItem;
	}

	/**
	 * @param element
	 * @throws JBIDescriptorException
	 */
	private ServiceAssembly loadServiceAssemblyXml(final Node node)
			throws JBIDescriptorException {
		ServiceAssembly serviceAssembly = new ServiceAssembly();

		Extensions extensions = new Extensions();
		// Services assembly sub nodes
		List<Node> children = XMLUtil.getNodeChildren(node);
		for (Node child : children) {
			if (IDENTIFICATION_ELEM_NAME.equals(child.getNodeName())) {
				serviceAssembly.setIdentification(loadIdentificationXml(child));
			} else if ("service-unit".equals(child.getNodeName())) {
				serviceAssembly.addServiceUnit(loadServiceUnitXml(child));
			} else if ("connections".equals(child.getNodeName())) {
				serviceAssembly.setConnections(loadConnectionsXml(child));
			} else {
				loadExtensions(extensions, child);
			}
		}
		// set extensions
		serviceAssembly.setExtensions(extensions);

		return serviceAssembly;
	}

	/**
	 * @param element
	 * @throws JBIDescriptorException
	 */
	private Services loadServicesXml(final Node production)
			throws JBIDescriptorException {

		Services services = new Services();

		// services attributes
		// is binding component (relative to service unit)
		String isBindingComponent;
		try {
			isBindingComponent = XMLUtil.getRequiredAttributeValue(production,
					"binding-component");
		} catch (NullPointerException e) {
			throw new JBIDescriptorException(e);
		}
		services.setBindingComponent(Boolean.parseBoolean(isBindingComponent));

		Extensions extensions = new Extensions();
		// Services sub nodes
		List<Node> children = XMLUtil.getNodeChildren(production);
		for (Node child : children) {
			if ("provides".equals(child.getNodeName())) {
				services.addProvides(loadProvidesXml(child));
			} else if ("consumes".equals(child.getNodeName())) {
				services.addConsumes(loadConsumesXml(child));
			} else {
				loadExtensions(extensions, child);
			}
		}
		// set extensions
		services.setExtensions(extensions);

		return services;
	}

	/**
	 * @param list
	 * @return
	 * @throws JBIDescriptorException
	 */
	private ServiceUnit loadServiceUnitXml(final Node node)
			throws JBIDescriptorException {

		ServiceUnit result = new ServiceUnit();

		Extensions extensions = new Extensions();
		// Service unit sub nodes
		List<Node> children = XMLUtil.getNodeChildren(node);
		for (Node child : children) {
			if (IDENTIFICATION_ELEM_NAME.equals(child.getNodeName())) {
				result.setIdentification(loadIdentificationXml(child));
			} else if ("target".equals(child.getNodeName())) {
				List<Node> childChildren = XMLUtil.getNodeChildren(child);
				for (Node targetChild : childChildren) {
					if ("artifacts-zip".equals(targetChild.getNodeName())) {
						result.setTargetArtifactsZip(XMLUtil
								.getTextContent(targetChild));
					} else if ("component-name".equals(targetChild
							.getNodeName())) {
						result.setTargetComponentName(XMLUtil
								.getTextContent(targetChild));
					}
				}
			} else {
				loadExtensions(extensions, child);
			}
		}
		// set extensions
		result.setExtensions(extensions);

		return result;
	}

	/**
	 * @param list
	 * @return
	 */
	private void loadSharedLibraryListXml(final List<SharedLibraryList> slList,
			final Node node) {
		SharedLibraryList sharedLibrary = new SharedLibraryList();
		sharedLibrary.setName(XMLUtil.getTextContent(node));
		sharedLibrary.setVersion(XMLUtil.getAttributeValue(node,
				VERSION_ATTR_NAME));
		slList.add(sharedLibrary);
	}

	/**
	 * @param element
	 * @throws JBIDescriptorException
	 */
	private SharedLibrary loadSharedLibraryXml(final Node production)
			throws JBIDescriptorException {
		SharedLibrary sharedLibrary = new SharedLibrary();

		// shared library attributes
		sharedLibrary.setClassLoaderDelegation(XMLUtil.getAttributeValue(
				production, "class-loader-delegation"));
		sharedLibrary.setVersion(XMLUtil.getAttributeValue(production,
				VERSION_ATTR_NAME));

		List<Node> children = XMLUtil.getNodeChildren(production);
		for (Node child : children) {
			if (IDENTIFICATION_ELEM_NAME.equals(child.getNodeName())) {
				sharedLibrary.setIdentification(loadIdentificationXml(child));
			} else if ("shared-library-class-path".equals(child.getNodeName())) {
				sharedLibrary.setSharedLibraryClassPath(XMLUtil
						.getTextContents(child.getChildNodes()));
			}
		}
		return sharedLibrary;
	}

	/**
	 * Load the JBISchema from the xsd file
	 * @return
	 * @throws JBIDescriptorException on SAXException or IOException
	 */
	protected Schema loadJBISchema() throws JBIDescriptorException { 

		if (schema == null) {
			// create a SchemaFactory capable of understanding WXS schemas
			SchemaFactory factory = SchemaFactory
					.newInstance(XMLConstants.W3C_XML_SCHEMA_NS_URI);

			// load a WXS schema, represented by a Schema instance
				try {
					schema = factory
							.newSchema(new StreamSource(openJbiSchemaFile()));
				} catch (SAXException e) {
					throw new JBIDescriptorException(e);
				} catch (IOException e) {
					throw new JBIDescriptorException(e);				}
		}
		return schema;
	}

	/**
	 * NOTE: The first location to search for JBI schema files is within the
	 * Petals installation root directory. If a schema file is not found within
	 * root directory, the alternative is use the packaged schema files flatten
	 * somewhere which can be found in class path. It won't work if the schema
	 * are compressed in a Zip or Jar archive.
	 *
	 * @return JBI schema as InputStream
	 * @throws IOException
	 *             if an error occurs during InputStream creation from schema
	 *             file
	 */
	private InputStream openJbiSchemaFile() throws IOException {
		InputStream theJbiSchemaResource = null;
		// search through classpath obtains from this class classloader
        URL schemaURL = this.getClass().getResource("/" + JBI_XSD);
        
		if (schemaURL != null) {
			theJbiSchemaResource = schemaURL.openStream();
		}

		if (theJbiSchemaResource == null) {
			// search in petals installation root
			String petalsRoot = System.getProperty("petals.home");
			if (petalsRoot != null) {
				theJbiSchemaResource = new FileInputStream(new File(petalsRoot
						+ JBI_SCHEMA_RESOURCE));
			}
		}
		return theJbiSchemaResource;
	}

}
